# -*- coding: utf-8 -*-
"""
Created on Wed Sep 22 17:22:59 2021

Default paths for the different folders.
Should be changed prior to any computation.

@author: amarmore
"""
# RWC
path_data_persisted_rwc = "C:/Users/amarmore/Desktop/data_persisted" ## Path where pre-computed data on RWC Pop should be stored (spectrograms, NTD, neural networks, etc)
path_annotation_rwc = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/annotations" ## Path of the annotations of RWC Pop. Should be parent folder of "AIST" and/or "MIREX10"
path_entire_rwc = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/Entire RWC" ## Path where are stored wav files of RWC
path_even_songs_rwc = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/Even songs" ## Path containing only the wav files of songs of odd numbers
path_odd_songs_rwc = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/Odd songs" ## Path containing only the wav files of songs of even numbers
path_debug_rwc = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/debug" ## Debug path, containing only two songs, to fix functions quickly
path_mirdata_rwc = "C:/Users/amarmore/Desktop/Audio samples/mirdata/RWC Pop/" ## Path of the annotations of RWC Pop. Should be parent folder of "AIST" and/or "MIREX10"

# SALAMI
path_data_persisted_salami = "C:/Users/amarmore/Desktop/Audio samples/Salami/data_persisted" ## Path where pre-computed data on the SALAMI dataset should be stored (spectrograms, NTD, neural networks, etc)
path_entire_salami = "C:/Users/amarmore/Desktop/Audio samples/Salami" ## Path where are stored wav files of SALAMI (path where it is downloaded by mirdata also)

# Come Together
come_together = "C:/Users/amarmore/this_folder/The Beatles - Come Together"
path_data_persisted_come_together = "C:/Users/amarmore/Desktop/data_persisted"